#include <Xm/Xm.h>
#include <Xi/Xi.h>
#include <Xi/TabList.h>

typedef struct _XiTabListRec {
    int    allocated;
    int    used;
    XiTabAttributes tabs;
} XiTabListRec;

#define XiLIST_GROWTH_FACTOR 10
#define XiXmStringCopy(s) (((s) == (XmString)NULL) ? NULL : XmStringCopy(s))
#define XiXmStringCompare(s1,s2) (((s1) == (s2)) ? True \
				  : ((s1) != NULL && (s2) != NULL) \
				  ? XmStringCompare(s1,s2) : False)
#define XiXmStringFree(s) if((s)!=(XmString)NULL)XmStringFree(s);else(s)=NULL

/*
 * Function:
 *      TabListCreate(void)
 * Description:
 *      Allocates and initializes an empty XiTabList structure and returns
 *      a pointer to the structure.
 * Input:
 *      None.
 * Output:
 *      XiTabList - an empty allocated XiTablist
 */
XiTabList
#ifndef _NO_PROTO
XiTabListCreate(void)
#else
XiTabListCreate()
#endif
{
    XiTabList tabList;

    /*
     * Allocate the structure and zero it.  Then give it back to the
     * caller.
     */
    tabList = (XiTabList) XtMalloc(sizeof(XiTabListRec));
    tabList->allocated = tabList->used = 0;
    tabList->tabs = (XiTabAttributes) NULL;

    return( tabList );
}

/*
 * Function:
 *      XiTabListCopy(tab_list)
 * Description:
 *      This makes a copy of the given XiTabList and returns the
 *      copy.
 * Input:
 *      tab_list : XiTabList - the XiTabList to copy
 * Output:
 *      XiTabList - the copy of the given list
 */
XiTabList
#ifndef _NO_PROTO
XiTabListCopy(XiTabList tab_list)
#else
XiTabListCopy(tab_list)
    XiTabList tab_list;
#endif
{
    XiTabList newList;
    int       i;

    /*
     * First if they gave us nothing, then lets return nothing to
     * them.
     */
    if( tab_list == NULL ) return( NULL );

    /*
     * Now allocate the new structure and set its allocated and used
     * to the number of items we need.
     */
    newList = (XiTabList) XtMalloc(sizeof(XiTabListRec));
    newList->allocated = newList->used = tab_list->used;

    /*
     * Now check if we have any items and if not just zero the
     * data, else we will have to copy item by item.
     */
    if( newList->used == 0 )
    {
        newList->tabs = (XiTabAttributes) NULL;
    }
    else
    {
        /*
         * First allocate the memory for all the entries. and then copy
         * item by item the list.
         */
        newList->tabs = (XiTabAttributes)
            XtMalloc(sizeof(XiTabAttributeRec) * newList->used);

        for( i = 0; i < newList->used; ++i )
        {
            newList->tabs[i].label_string =
                XiXmStringCopy(tab_list->tabs[i].label_string);
            newList->tabs[i].label_pixmap = tab_list->tabs[i].label_pixmap;
	    newList->tabs[i].string_direction =
		tab_list->tabs[i].string_direction;
	    newList->tabs[i].pixmap_placement =
		tab_list->tabs[i].pixmap_placement;
	    newList->tabs[i].label_alignment =
		tab_list->tabs[i].label_alignment;
            newList->tabs[i].foreground = tab_list->tabs[i].foreground;
            newList->tabs[i].background = tab_list->tabs[i].background;
	    newList->tabs[i].background_pixmap =
		tab_list->tabs[i].background_pixmap;
	    newList->tabs[i].sensitive = tab_list->tabs[i].sensitive;
            newList->tabs[i].value_mode = XiTAB_VALUE_COPY;
        }
    }

    return( newList );
}

/*
 * Function:
 *      XiTabListFree(tab_list)
 * Description:
 *      Deallocates the memory associated with the given tab list.
 * Input:
 *      tab_list : XiTablist - the XiTabList to deallocate
 * Output:
 *      None.
 */
void
#ifndef _NO_PROTO
XiTabListFree(XiTabList tab_list)
#else
XiTabListFree(tab_list)
    XiTabList tab_list;
#endif
{
    int i;

    /*
     * If we do not have anything to deallocate the lets just
     * get out of here.
     */
    if( tab_list == NULL ) return;

    /*
     * If we have used items then lets walk through the list and
     * deallocate anything we have to.
     */
    if( tab_list->used != 0 )
    {
        for( i = 0; i < tab_list->used; ++i )
        {
            if( tab_list->tabs[i].value_mode == XiTAB_VALUE_SHARE ) continue;
            
            XiXmStringFree(tab_list->tabs[i].label_string);
        }
    }

    /*
     * If we have any allocated slots lets deallocate them and finally
     * deallocate the parent structure.
     */
    if( tab_list->allocated != 0 )
    {
        XtFree((XtPointer) tab_list->tabs);
    }
    XtFree((XtPointer) tab_list);
}

/*
 * Function:
 *      XiTabListRemove(tab_list, position)
 * Description:
 *      Remove and deallocate the tab at the specified position in
 *      the list.
 * Input:
 *      tab_list : XiTabList - the XiTabList to work with
 *      position : int       - the position of the tab to remove
 * Output:
 *      None.
 */
void
#ifndef _NO_PROTO
XiTabListRemove(XiTabList tab_list, int position)
#else
XiTabListRemove(tab_list, position)
    XiTabList tab_list;
    int       position;
#endif
{
    XiTabAttributes to, from;
    int             count;

    if( tab_list == NULL || position < 0 ||
        (position > 0 && position >= tab_list->used) )
    {
        return;
    }
    to = &(tab_list->tabs[position]);

    /*
     * Now deallocate the data associated with this tab, if needed.
     */
    if( to->value_mode == XiTAB_VALUE_COPY ) XiXmStringFree(to->label_string);

    /*
     * Now decrement the number of used slots and shift down any slots
     * above the one that is being deleted.
     */
    tab_list->used--;
    if( position != tab_list->used )
    {
        from = &(tab_list->tabs[position+1]);
        count = tab_list->used - position;
        memmove((XtPointer) to, (XtPointer) from,
                sizeof(XiTabAttributeRec) * count);
    }
}

/*
 * Function:
 *      XiTabListInsert(tab_list, position, mask, attributes)
 * Description:
 *      Adds a tab with the specified attributes to the given
 *      XiTabList.
 * Input:
 *      tab_list   : XiTabList       - the XiTabList to add the new tab to
 *      position   : int             - the position to place the new tab at
 *      mask       : XtValueMask     - the mask of attributes to set on the
 *                                     new tab
 *      attributes : XiTabAttributes - the list of attributes to set on the
 *				       new tab
 * Output:
 *      int - the position of the new tab in the list
 */
int
#ifndef _NO_PROTO
XiTabListInsert(XiTabList tab_list, int position, XtValueMask mask,
                XiTabAttributes attributes)
#else
XiTabListInsert(tab_list, position, mask, attributes)
    XiTabList       tab_list;
    int             position;
    XtValueMask     mask;
    XiTabAttributes attributes;
#endif
{
    XiTabAttributes newTab;

    /*
     * First lets see if they gave us a valid XiTabList and position
     * and if not lets just leave.
     */
    if( tab_list == NULL || position < XiTAB_LAST_POSITION ||
        (position > 0 && position >= tab_list->used) ) return;

    /*
     * Now lets add our new tab to the list in the correct position.
     * If the list is not large enough to hold our new tab we will 
     * grow the list.
     */
    if( tab_list->allocated == tab_list->used )
    {
	/*
	 * Our list is not large enough to take another tab, so lets
	 * grow the list by XiLIST_GROWTH_FACTOR.
	 */
	tab_list->allocated += XiLIST_GROWTH_FACTOR;
	tab_list->tabs = (XiTabAttributes)
	    XtRealloc((XtPointer) tab_list->tabs,
		      sizeof(XiTabAttributeRec) * tab_list->allocated);
    }

    /*
     * Once here we know that the list is large enough to hold our new
     * item so now lets move any existing items up to make room for our
     * new kiddie.
     */
    if( position == XiTAB_LAST_POSITION )
    {
	newTab = &(tab_list->tabs[tab_list->used]);
	position = tab_list->used;
    }
    else
    {
	XiTabAttributes to, from;
	int             count;

	from = &(tab_list->tabs[position]);
	to = &(tab_list->tabs[position+1]);
	count = tab_list->used - position;
	memmove(to, from, sizeof(XiTabAttributeRec) * count);
	newTab = &(tab_list->tabs[position]);
    }
    tab_list->used++;

    /*
     * Now lets initialize the new tab.  After that we will take a look
     * at the mask and see what attributes were customized.
     */
    newTab->label_string = (XmString) NULL;
    newTab->string_direction = XmSTRING_DIRECTION_L_TO_R;
    newTab->label_pixmap = XmUNSPECIFIED_PIXMAP;
    newTab->label_alignment = XmALIGNMENT_CENTER;
    newTab->pixmap_placement = XiPIXMAP_RIGHT;
    newTab->foreground = XiCOLOR_DYNAMIC;
    newTab->background = XiCOLOR_DYNAMIC;
    newTab->background_pixmap = XiPIXMAP_DYNAMIC;
    newTab->sensitive = True;
    newTab->value_mode = XiTAB_VALUE_COPY;

    if( mask & XiTAB_VALUE_MODE )
    {
	newTab->value_mode = attributes->value_mode;
    }

    if( mask & XiTAB_LABEL_STRING )
    {
	if( newTab->value_mode == XiTAB_VALUE_COPY )
	{
	    newTab->label_string = XiXmStringCopy(attributes->label_string);
	}
	else
	{
	    newTab->label_string = attributes->label_string;
	}
    }
    
    if( mask & XiTAB_STRING_DIRECTION )
    {
	newTab->string_direction = attributes->string_direction;
    }

    if( mask & XiTAB_LABEL_PIXMAP )
    {
	newTab->label_pixmap = attributes->label_pixmap;
    }

    if( mask & XiTAB_PIXMAP_PLACEMENT )
    {
	newTab->pixmap_placement = attributes->pixmap_placement;
    }

    if( mask & XiTAB_BACKGROUND )
    {
	newTab->background = attributes->background;
    }

    if( mask & XiTAB_BACKGROUND_PIXMAP )
    {
	newTab->background_pixmap = attributes->background_pixmap;
    }

    if( mask * XiTAB_SENSITIVE )
    {
	newTab->sensitive = attributes->sensitive;
    }

    if( mask & XiTAB_FOREGROUND )
    {
	newTab->foreground = attributes->foreground;
    }
    
    if( mask & XiTAB_LABEL_ALIGNMENT )
    {
	newTab->label_alignment = attributes->label_alignment;
    }

    return( position );
}

/*
 * Function:
 *	XiTabListAppend(tab_list, mask, attributes)
 * Description:
 *	Appends a new tab with the given attributes.
 * Input:
 *	tab_list   : XiTabList       - the XiTabList to append a tab to
 *	mask       : XtValueMask     - the mask of attributes to set on the
 *				       new tab
 *	attributes : XiTabAttributes - the list of attributes to set on the
 *				       new tab
 * Output:
 *	int - the position of the new tab
 */
int
#ifndef _NO_PROTO
XiTabListAppend(XiTabList tab_list, XtValueMask mask,
                XiTabAttributes attributes)
#else
XiTabListAppend(tab_list, mask, attributes)
    XiTabList       tab_list;
    XtValueMask     mask;
    XiTabAttributes attributes;
#endif
{
    /*
     * So we cheat a little here.  All we need to do is call insert and
     * lets that function know that we want it at the end of the list.
     */
    return( XiTabListInsert(tab_list, XiTAB_LAST_POSITION, mask, attributes) );
}

/*
 * Function:
 *	XiTabListModify(tab_list, position, mask, attributes)
 * Description:
 *	Change the specified attributes for the tab at the given position
 *	in the given XiTabList.
 * Input:
 *	tab_list   : XiTabList       - the XiTabList that contains the
 *				       tab to modify.
 *	position   : int             - the index of the tab to modify
 *	mask       : XtValueMask     - the mask of attributes to modify
 *	attributes : XiTabAttributes - the list of attributes to modify
 * Output:
 *	None.
 */
void
#ifndef _NO_PROTO
XiTabListModify(XiTabList tab_list, int position, XtValueMask mask,
                XiTabAttributes attributes)
#else
XiTabListModify(tab_list, position, mask, attributes)
    XiTabList       tab_list;
    int             position;
    XtValueMask     mask;
    XiTabAttributes attributes;
#endif
{
    XiTabAttributes tab;

    /*
     * First lets see if they gave us anything valid to work with and
     * if they did not lets vacate the premises.
     */
    if( tab_list == NULL || position < 0 || position >= tab_list->used )
    {
	return;
    }

    tab = &(tab_list->tabs[position]);

    /*
     * Now lets check the various flags and see what we need to 
     * change.
     */
    if( mask & XiTAB_VALUE_MODE && tab->value_mode != attributes->value_mode &&
        attributes->value_mode == XiTAB_VALUE_COPY )
    {
	/*
	 * Can only change this mode from share to copy, NOT from copy
	 * to share.
	 */
	tab->value_mode = attributes->value_mode;
	tab->label_string = XiXmStringCopy(attributes->label_string);
    }
    
    if( mask & XiTAB_LABEL_STRING )
    {
	if( tab->value_mode == XiTAB_VALUE_COPY )
	{
	    XiXmStringFree(tab->label_string);
	    tab->label_string = XiXmStringCopy(attributes->label_string);
	}
	else
	{
	    tab->label_string = attributes->label_string;
	}
    }

    if( mask & XiTAB_STRING_DIRECTION )
    {
	tab->string_direction = attributes->string_direction;
    }

    if( mask & XiTAB_LABEL_PIXMAP )
    {
	tab->label_pixmap = attributes->label_pixmap;
    }

    if( mask & XiTAB_PIXMAP_PLACEMENT )
    {
	tab->pixmap_placement = attributes->pixmap_placement;
    }

    if( mask & XiTAB_BACKGROUND )
    {
	tab->background = attributes->background;
    }

    if( mask & XiTAB_BACKGROUND_PIXMAP )
    {
	tab->background_pixmap = attributes->background_pixmap;
    }

    if( mask & XiTAB_SENSITIVE )
    {
	tab->sensitive = attributes->sensitive;
    }

    if( mask & XiTAB_FOREGROUND )
    {
	tab->foreground = attributes->foreground;
    }

    if( mask & XiTAB_LABEL_ALIGNMENT )
    {
	tab->label_alignment = attributes->label_alignment;
    }
}

/*
 * Function:
 *	XiTabListQuery(tab_list, position, mask, attributes)
 * Description:
 *	Retrieve attributes from the specified tab. (NOTE: This may be 
 *	a copy out, depends on value_mode, so be sure to free
 *	this when done.)
 * Input:
 *	tab_list   : XiTabList       - the tab list that contains the tab
 *	position   : int             - the tab index to query
 *	attributes : XiTabAttributes - the attributes returned.
 * Output:
 *	None.
 */
void
#ifndef _NO_PROTO
XiTabListQuery(XiTabList tab_list, int position, XiTabAttributes attributes)
#else
XiTabListQuery(tab_list, position, attributes)
    XiTabList       tab_list;
    int             position;
    XiTabAttributes attributes;
#endif
{
    XiTabAttributes tab;

    /*
     * Make sure that we have valid data and if not "asta la vesta, baby"
     * I have NO idea how to spell that.
     */
    if( tab_list == NULL || position < 0 || position >= tab_list->used )
    {
	return;
    }

    tab = &(tab_list->tabs[position]);

    if( tab->value_mode == XiTAB_VALUE_COPY )
    {
	attributes->label_string = XiXmStringCopy(tab->label_string);
    }
    else
    {
	attributes->label_string = tab->label_string;
    }

    attributes->label_pixmap = tab->label_pixmap;
    attributes->string_direction = tab->string_direction;
    attributes->label_alignment = tab->label_alignment;
    attributes->pixmap_placement = tab->pixmap_placement;
    attributes->background = tab->background;
    attributes->background_pixmap = tab->background_pixmap;
    attributes->sensitive = tab->sensitive;
    attributes->foreground = tab->foreground;
    attributes->value_mode = tab->value_mode;
}

/*
 * Function:
 *	XiTabListFind(tab_list, label_string)
 * Description:
 *	Finds the first tab whose label_string matches the given
 *	and returns its index.
 * Input:
 *	tab_list     : XiTabList - the XiTabList to search
 *	label_string : XmString  - the label_string to search for
 * Output:
 *	int - the index of the tab found, or XiTAB_NOT_FOUND if 
 *	      a match is not found.
 */
int
#ifndef _NO_PROTO
XiTabListFind(XiTabList tab_list, XmString label_string)
#else
XiTabListFind(tab_list, label_string)
    XiTabList tab_list;
    XmString  label_string;
#endif
{
    int i;

    if( tab_list == NULL ) return( XiTAB_NOT_FOUND );

    for( i = 0; i < tab_list->used; ++i )
    {
	if( XiXmStringCompare(tab_list->tabs[i].label_string, label_string) )
	{
	    return( i );
	}
    }
    
    return( XiTAB_NOT_FOUND );
}

/*
 * Function:
 *	XiTabListSimpleRemove(tab_list, label_string)
 * Description:
 *	Removes the first tab from the given XiTabList that matches the
 *	given label string.
 * Input:
 *	tab_list     : XiTabList - tab list to remove the item from
 *	label_string : XmString  - the label string to match
 * Output:
 *	None.
 */
void
#ifndef _NO_PROTO
XiTabListSimpleRemove(XiTabList tab_list, XmString label_string)
#else
XiTabListSimpleRemove(tab_list, label_string)
    XiTabList tab_list;
    XmString  label_string;
#endif
{
    int position = XiTabListFind(tab_list, label_string);

    if( position != XiTAB_NOT_FOUND ) XiTabListRemove(tab_list, position);
}

/*
 * Function:
 *	XiTabListSimpleInsert(tab_list, position, label_string)
 * Description:
 *	Inserts a new tab with the specified label string at the
 *	specified position.
 * Input:
 *	tab_list     : XiTabList - the XiTabList to add a tab to
 *	position     : int       - the position to put the tab at
 *	label_string : XmString  - the label for the new tab
 * Output:
 *	int - the position of the new tab
 */
int
#ifndef _NO_PROTO
XiTabListSimpleInsert(XiTabList tab_list, int position, XmString label_string)
#else
XiTabListSimpleInsert(tab_list, position, label_string)
    XiTabList tab_list;
    int       position;
    XmString  label_string;
#endif
{
    XiTabAttributeRec attributes;

    attributes.label_string = label_string;
    
    return( XiTabListInsert(tab_list, position, XiTAB_LABEL_STRING,
			    &attributes) );
}

/*
 * Function:
 *	XiTabListSimpleAppend(tab_list, label_string)
 * Description:
 *	Append a new tab to the given tab list with the specified label
 *	string.
 * Input:
 *	tab_list     : XiTabList - the XiTabList to append a tab to
 *	label_string : XmString  - the label for the new tab
 * Output:
 *	int - the position of the new tab
 */
int
#ifndef _NO_PROTO
XiTabListSimpleAppend(XiTabList tab_list, XmString label_string)
#else
XiTabListSimpleAppend(tab_list, label_string)
    XiTabList tab_list;
    XmString  label_string;
#endif
{
    XiTabAttributeRec attributes;

    attributes.label_string = label_string;
    
    return( XiTabListAppend(tab_list, XiTAB_LABEL_STRING,
			    &attributes) );
}

/*
 * Function:
 *	XiTabListsimpleModify(tab_list, position, label_string)
 * Description:
 *	Changes the label string for the specified tab.
 * Input:
 *	tab_list     : XiTabList - the XiTabList that contains the tab to
 *				   change
 *	position     : int       - the tab to change
 * 	label_string : XmString  - the new value of the label string
 * Output:
 *	None.
 */
void
#ifndef _NO_PROTO
XiTabListSimpleModify(XiTabList tab_list, int position, XmString label_string)
#else
XiTabListSimpleModify(tab_list, position, label_string)
    XiTabList tab_list;
    int       position;
    XmString  label_string;
#endif
{
    XiTabAttributeRec attributes;

    XiTabListModify(tab_list, position, XiTAB_LABEL_STRING, &attributes);
}

/*
 * Function:
 *	XiTabListSimpleQuery(tab_list, position)
 * Description:
 *	Query the label string from the specified tab
 * Input:
 *	tab_list : XiTabList - the XiTabList to query
 *	position : int       - the tab to query
 * Output:
 *	XmString - the label for the given tab (NOTE: this is a copy
 *	           so please free when you are done)
 */
XmString
#ifndef _NO_PROTO
XiTabListSimpleQuery(XiTabList tab_list, int position)
#else
XiTabListSimpleQuery(tab_list, position)
    XiTabList tab_list;
    int       position;
#endif
{
    XiTabAttributeRec attributes;

    XiTabListQuery(tab_list, position, &attributes);

    if( attributes.value_mode != XiTAB_VALUE_COPY )
    {
	attributes.label_string = XiXmStringCopy(attributes.label_string);
    }

    return( attributes.label_string );
}

/*
 * Function:
 *	XiTabAttributesFree(attributes)
 * Description:
 *	This deallocates any memory that the attribute structure is
 *	using.  This does not free the attribute structure.
 * Input:
 *	attributes : XiAttributes - the structure whose contants should be
 *		                    deallocated.
 * Output:
 *	None.
 */
void
#ifndef _NO_PROTO
XiTabAttributesFree(XiTabAttributes attributes)
#else
XiTabAttributesFree(attributes)
    XiTabAttributes attributes;
#endif
{
    if( attributes->value_mode == XiTAB_VALUE_COPY )
    {
	XiXmStringFree(attributes->label_string);
    }
}

/*
 * Function:
 *	_XiTabListCount(tab_list)
 * Description:
 *	Returns the number of entries in the given XiTabList.
 * Input:
 *	tab_list : XiTabList - the tab list whose entries to count
 * Output:
 *	int - the number of items in the given XiTabList
 */
int
#ifndef _NO_PROTO
_XiTabListCount(XiTabList tab_list)
#else
_XiTabListCount(tab_list)
    XiTabList tab_list;
#endif
{
    if( tab_list == NULL ) return( 0 );

    return( tab_list->used );
}

/*
 * Function:
 *	_XiTabListGet(tab_list, position)
 * Description:
 *	Returns a pointer to the given element in the XiTabList.  Use
 *	with caution as it returns a pointer to live memory.
 * Input:
 *	tab_list : XiTabList - the tab list to index into
 *	int      : Position  - the index of the tab you want
 * Output:
 *	XiTabAttributes - a pointer to the specified tab
 */
XiTabAttributes
#ifndef _NO_PROTO
_XiTabListGet(XiTabList tab_list, int position)
#else
_XiTabListGet(tab_list, position)
    XiTabList tab_list;
    int       position;
#endif
{
    XiTabAttributes tab;

    /*
     * Make sure that we have valid data and if not "asta la vesta, baby"
     * I have NO idea how to spell that.
     */
    if( tab_list == NULL || position < 0 || position >= tab_list->used )
    {
	return( NULL );
    }

    tab = &(tab_list->tabs[position]);

    return( tab );
}

XiTabAttributes
#ifndef _NO_PROTO
_XiTabListArray(XiTabList tab_list)
#else
_XiTabListArray(tab_list)
    XiTabList tab_list;
#endif
{
    /*
     * Make sure that we have valid data and if not "asta la vesta, baby"
     * I have NO idea how to spell that.
     */
    if( tab_list == NULL )
    {
	return( NULL );
    }

    return( tab_list->tabs );
}

/*
 * Function:
 *	XiTabListCompare(XiTabList list1, XiTabList list2)
 * Description:
 *	Compares to tab lists to see if they are logically 
 *	equivalent.
 * Input:
 *	list1 : XiTabList - tab list to compare
 *	list2 : XiTabList - tab list to compare
 * Output:
 *	XiTabResult - indicates the differences it any
 */
XiTabResult
#ifndef _NO_PROTO
XiTabListCompare(XiTabList list1, XiTabList list2)
#else
XiTabListCompare(list1, list2)
    XiTabList list1, list2;
#endif
{
    XiTabResult result = XiTAB_CMP_EQUAL;
    int         i;

    if( list1 == NULL && list2 != NULL ||
        list1 != NULL && list2 == NULL ||
        list1->used != list2->used ) return( XiTAB_CMP_SIZE );

    for( i = 0; i < list1->used; ++i )
    {
	if( list1->tabs[i].label_pixmap != list2->tabs[i].label_pixmap ||
	    !XiXmStringCompare(list1->tabs[i].label_string, 
			       list2->tabs[i].label_string) ||
 	    list1->tabs[i].pixmap_placement !=
	    list2->tabs[i].pixmap_placement )
	{
	    return( XiTAB_CMP_SIZE );
	}
	if( list1->tabs[i].string_direction !=
	    list2->tabs[i].string_direction ||
	    list1->tabs[i].label_alignment != list2->tabs[i].label_alignment ||
	    list1->tabs[i].foreground != list2->tabs[i].foreground ||
	    list1->tabs[i].background != list2->tabs[i].background ||
	    list1->tabs[i].sensitive != list2->tabs[i].sensitive ||
	    list1->tabs[i].background_pixmap != 
	    list2->tabs[i].background_pixmap )
	{
	    result = XiTAB_CMP_VISUAL;
	}
    }
    return( result );
}
